use std::ops::DerefMut;

use maybe_owned::MaybeOwnedMut;
use url::Url;

use crate::{client::Client, common::*, error, response::PagedResponse};

type Parameter<'a> = (&'a str, &'a str);

const RELEASE_API_URL: &'static str = "https://services.api.unity.com/unity/editor/release/";

pub struct RequestBuilder<'a, T> {
    client: Option<MaybeOwnedMut<'a, T>>,
    url: Option<Url>,
    api_version: APIVersion,
    limit: u8,
    offset: usize,
    order: Order,
    stream: Option<Stream>,
    platform: Option<Platform>,
    architecture: Option<Architecture>,
    version: Option<Version>,
}

impl<T: Default> Default for RequestBuilder<'_, T> {
    fn default() -> Self {
        Self {
            limit: 10,
            client: Default::default(),
            url: Some(Self::default_api_url()),
            api_version: Default::default(),
            offset: Default::default(),
            order: Default::default(),
            stream: Default::default(),
            platform: Default::default(),
            architecture: Default::default(),
            version: Default::default(),
        }
    }
}

impl<'a, T: 'a> RequestBuilder<'a, T> {
    fn default_api_url() -> Url {
        Url::parse(RELEASE_API_URL).unwrap()
    }
}

impl<'a, T: 'a + Client> RequestBuilder<'a, T> {
    pub fn new<I: Into<MaybeOwnedMut<'a, T>>>(client: I) -> RequestBuilder<'a, T> {
        Self {
            client: Some(client.into()),
            url: Some(Self::default_api_url()),
            limit: 10,
            api_version: Default::default(),
            offset: Default::default(),
            order: Default::default(),
            stream: Default::default(),
            platform: Default::default(),
            architecture: Default::default(),
            version: Default::default(),
        }
    }

    pub fn with_client<I: Into<MaybeOwnedMut<'a, T>>>(self, client: I) -> RequestBuilder<'a, T> {
        Self {
            client: Some(client.into()),
            ..self
        }
    }

    pub fn with_url(self, url: Url) -> Self {
        Self {
            url: Some(url),
            ..self
        }
    }

    pub fn with_limit(self, limit: u8) -> Self {
        Self { limit, ..self }
    }

    pub fn with_offset(self, offset: usize) -> Self {
        Self { offset, ..self }
    }

    pub fn with_order(self, order: Order) -> Self {
        Self { order, ..self }
    }

    pub fn with_stream(self, stream: Stream) -> Self {
        Self {
            stream: Some(stream),
            ..self
        }
    }

    pub fn with_platform(self, platform: Platform) -> Self {
        Self {
            platform: Some(platform),
            ..self
        }
    }

    pub fn with_architecture(self, architecture: Architecture) -> Self {
        Self {
            architecture: Some(architecture),
            ..self
        }
    }

    pub fn with_version(self, version: String) -> Self {
        Self {
            version: Some(Version(version)),
            ..self
        }
    }

    pub async fn send(self) -> Result<PagedResponse, error::Error> {
        Into::<Request<'a, T>>::into(self).send().await
    }
}

impl<'a, T: 'a + Client> From<RequestBuilder<'a, T>> for Request<'a, T> {
    fn from(val: RequestBuilder<'a, T>) -> Self {
        Request {
            client: val.client.unwrap(),
            api_url: val.url.unwrap_or_else(RequestBuilder::<T>::default_api_url),
            api_version: val.api_version.to_string(),
            limit: val.limit.to_string(),
            offset: val.offset.to_string(),
            order: val.order.to_string(),
            stream: val.stream.map_or_else(String::default, |s| s.to_string()),

            platform: val.platform.map_or_else(String::default, |p| p.to_string()),

            architecture: val
                .architecture
                .map_or_else(String::default, |a| a.to_string()),

            version: val
                .version
                .map_or_else(String::default, |v| v.0.to_string()),
        }
    }
}

pub struct Request<'a, T> {
    api_url: Url,
    client: MaybeOwnedMut<'a, T>,
    api_version: String,
    limit: String,
    offset: String,
    order: String,
    stream: String,
    platform: String,
    architecture: String,
    version: String,
}

impl<T: Client> Request<'_, T> {
    fn as_parameters(&self) -> impl Iterator<Item = Parameter<'_>> {
        [
            ("limit", self.limit.as_str()),
            ("offset", self.offset.as_str()),
            ("order", self.order.as_str()),
            ("stream", self.stream.as_str()),
            ("platform", self.platform.as_str()),
            ("architecture", self.architecture.as_str()),
            ("version", self.version.as_str()),
        ]
        .into_iter()
        .filter(|p| !p.1.is_empty())
    }

    fn get_url(&self) -> Result<Url, url::ParseError> {
        let version = format!("{}/", self.api_version);
        self.api_url.join(&version)?.join("releases")
    }

    fn build_uri(&self) -> Result<Url, url::ParseError> {
        self.get_url()
            .and_then(|url| Url::parse_with_params(url.as_str(), self.as_parameters()))
    }

    pub async fn send(&mut self) -> Result<PagedResponse, error::Error> {
        let url = self.build_uri().unwrap();
        let req = self.client.get(url)?;
        self.client.fetch(req).await
    }
}

#[derive(Default)]
pub struct UnityReleaseClient<'a, T> {
    client: MaybeOwnedMut<'a, T>,
}

impl<'a, T: 'a> UnityReleaseClient<'a, T> {
    pub fn new(client: impl Into<MaybeOwnedMut<'a, T>>) -> Self {
        Self {
            client: client.into(),
        }
    }
}

impl<'a, T: 'a + Client> UnityReleaseClient<'a, T> {
    pub fn request(&mut self) -> RequestBuilder<'_, T> {
        RequestBuilder::new(MaybeOwnedMut::Borrowed(self.client.deref_mut()))
    }

    pub async fn send(
        &self,
        request: impl Into<Request<'a, T>>,
    ) -> Result<PagedResponse, error::Error> {
        request.into().send().await
    }
}

#[cfg(test)]
mod tests {
    use crate::request::*;
    use reqwest;

    #[test]
    fn defaut_request_uri() {
        let expected = "https://services.api.unity.com/unity/editor/release/v1/releases?limit=10&offset=0&order=RELEASE_DATE_DESC";
        let mut client = UnityReleaseClient::new(reqwest::Client::default());
        let request: Request<reqwest::Client> = client.request().into();
        assert_eq!(request.build_uri().unwrap().as_str(), expected);
    }

    #[test]
    fn custom_request_uri() {
        let expected = "https://services.api.unity.com/unity/editor/release/v1/releases?limit=5&offset=20&order=RELEASE_DATE_ASC&stream=LTS&platform=LINUX&architecture=ARM64&version=2020.3.44f1";

        let mut client = UnityReleaseClient::default();
        let request: Request<reqwest::Client> = client
            .request()
            .with_order(Order::Ascending)
            .with_limit(5)
            .with_offset(20)
            .with_stream(Stream::Lts)
            .with_platform(Platform::Linux)
            .with_architecture(Architecture::Arm64)
            .with_version("2020.3.44f1".to_string())
            .into();

        assert_eq!(request.build_uri().unwrap().as_str(), expected);
    }

    #[test]
    fn borrowed_client() {
        let expected = "https://services.api.unity.com/unity/editor/release/v1/releases?limit=10&offset=0&order=RELEASE_DATE_ASC";
        let request_client = reqwest::Client::default();
        let mut client = UnityReleaseClient::new(request_client);
        let request: Request<reqwest::Client> =
            client.request().with_order(Order::Ascending).into();

        assert_eq!(request.build_uri().unwrap().as_str(), expected);
    }

    //#[tokio::test]
    //async fn test_connection() {
    //    let mut client = UnityReleaseClient::default();
    //    let request = client.request().with_offset(25).with_limit(25);
    //    let mut req: Request<reqwest::Client> = request.into();
    //    // println!("{}", req.build_uri().unwrap());
    //    let response = req.send().await;
    //    print!("{response:?}");
    //}
}
