use std::{fmt::Display, path::PathBuf, str::FromStr};

use serde::Deserialize;
use subresource_integrity::Integrity;
use time::Date;

use crate::{common::*, error::Error};

fn handle_fractional_numbers<'de, D>(deserializer: D) -> Result<usize, D::Error>
where
    D: serde::Deserializer<'de>,
{
    let value: f64 = Deserialize::deserialize(deserializer)?;
    let value = value.ceil();
    Ok(value as usize)
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "SCREAMING_SNAKE_CASE")]
pub enum UnitKind {
    Byte,
    Kilobyte,
    Megabyte,
    Gigabyte,
}

#[derive(Debug, serde::Deserialize)]
pub struct DigitalValue {
    #[serde(deserialize_with = "handle_fractional_numbers")]
    pub value: usize,
    pub unit: UnitKind,
}

#[derive(Debug, serde::Deserialize)]
pub struct ReleaseNotes {
    pub url: PathBuf,
    pub integrity: Option<Integrity>,
    #[serde(rename = "type")]
    pub kind: FileKind,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct ThirdPartyNotice {
    #[serde(rename = "originalFileName")]
    pub original_filename: String,
    pub url: PathBuf,
    pub integrity: Option<Integrity>,
    #[serde(rename = "type")]
    pub kind: FileKind,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "UPPERCASE")]
pub enum FileKind {
    Text,
    #[serde(rename = "TAR_GZ")]
    TarGz,
    #[serde(rename = "TAR_XZ")]
    TarXz,
    Zip,
    Pkg,
    Exe,
    Po,
    Dmg,
    Lzma,
    Lz4,
    Md,
    Pdf,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "SCREAMING_SNAKE_CASE")]
pub enum ModuleCategory {
    Documentation,
    Platform,
    LanguagePack,
    DevTool,
    Plugin,
    Component,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "SCREAMING_SNAKE_CASE")]
pub enum SKUFamily {
    Dots,
    Classic,
}

#[derive(Debug, serde::Deserialize)]
pub struct ExtractedPathRename {
    pub from: PathBuf,
    pub to: PathBuf,
}

#[derive(Debug, serde::Deserialize)]
pub struct Eula {
    pub url: PathBuf,
    pub integrity: Option<Integrity>,
    #[serde(rename = "type")]
    pub kind: FileKind,
    pub label: String,
    pub message: String,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Module {
    pub id: String,
    pub slug: String,
    pub name: String,
    pub description: String,
    pub category: ModuleCategory,
    pub url: PathBuf,
    #[serde(rename = "type")]
    pub kind: FileKind,
    pub download_size: DigitalValue,
    pub installed_size: DigitalValue,
    pub sub_modules: Vec<Module>,
    pub required: bool,
    pub hidden: bool,
    pub extracted_path_rename: Option<ExtractedPathRename>,
    #[serde(rename = "preSelected")]
    pub preselected: bool,
    pub destination: Option<String>,
    pub eula: Option<Vec<Eula>>,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Download {
    pub url: PathBuf,
    pub integrity: Option<Integrity>,
    #[serde(rename = "type")]
    pub kind: FileKind,
    pub platform: Platform,
    pub architecture: Architecture,
    pub download_size: DigitalValue,
    pub installed_size: DigitalValue,
    pub modules: Vec<Module>,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Release {
    pub version: Version,
    pub date: Option<Date>,
    pub notes: Option<ReleaseNotes>,
    pub stream: Stream,
    pub downloads: Vec<Download>,
    pub sku_family: SKUFamily,
    pub recommended: bool,
    pub unity_hub_link: Option<String>,
    pub short_revision: String,
    pub third_party_notices: Vec<ThirdPartyNotice>,
}

#[derive(Debug, serde::Deserialize)]
pub struct PagedResponse {
    pub offset: usize,
    pub limit: u8,
    pub total: usize,
    pub results: Vec<Release>,
}

#[derive(Debug, Deserialize)]
#[serde(untagged)]
pub enum APIResponse {
    Ok(PagedResponse),
    Err(APIError),
}

impl FromStr for PagedResponse {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        serde_json::from_str(s).map_err(Error::from)
    }
}

#[derive(Debug, serde::Deserialize)]
pub struct APIError {
    pub title: String,
    pub status: u16,
    pub detail: String,
}

impl Display for APIError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{} ({}): {}", self.title, self.status, self.detail)
    }
}

impl core::error::Error for APIError {}
